from .core import (
    update_q_from_qind,
    update_primal_from_dual,
    get_jacobian_and_residual,
    compute_jacobian,
    parallelise_edges,
)
from .graphstatics import (
    form_identify_dof,
    form_count_dof,
    form_update_q_from_qind,
    form_update_from_force,
    form_update_from_force_newton,
    force_update_from_form,
    force_update_from_constraints,
    update_diagrams_from_constraints,
)
from .loadpath import (
    compute_loadpath,
    compute_external_work,
    compute_internal_work,
    compute_internal_work_tension,
    compute_internal_work_compression,
    optimise_loadpath,
)
from .constraints import (
    ConstraintsCollection,
    HorizontalFix,
    VerticalFix,
    AngleFix,
    LengthFix,
    SetLength,
)

__all__ = [
    "update_q_from_qind",
    "update_primal_from_dual",
    "get_jacobian_and_residual",
    "compute_jacobian",
    "parallelise_edges",
    "form_identify_dof",
    "form_count_dof",
    "form_update_q_from_qind",
    "form_update_from_force",
    "form_update_from_force_newton",
    "force_update_from_form",
    "force_update_from_constraints",
    "update_diagrams_from_constraints",
    "compute_loadpath",
    "compute_external_work",
    "compute_internal_work",
    "compute_internal_work_tension",
    "compute_internal_work_compression",
    "optimise_loadpath",
    "ConstraintsCollection",
    "HorizontalFix",
    "VerticalFix",
    "AngleFix",
    "LengthFix",
    "SetLength",
]



from compas.datastructures import Graph


class FormGraph(Graph):
    """A graph representing the geometry and connectivity of the lines of a form diagram."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def node_index(self) -> dict[int, int]:
        return {node: index for index, node in enumerate(self.nodes())}

    def is_2d(self) -> bool:
        """Verify that all nodes of the graph lie in a horizontal plane.

        Returns
        -------
        bool

        """
        z = self.nodes_attribute("z")
        zmin = min(z)
        zmax = max(z)
        if (zmax - zmin) ** 2 > 0.001:
            return False
        return True

    def is_planar_embedding(self) -> bool:
        """Verify that the current embedding of the graph is planar."""
        return self.is_2d() and self.is_planar() and not self.is_crossed()

    def embed(self, fixed: list[int] = None, straightline: bool = True) -> bool:
        """Compute a geometry for the graph that embeds it in the plane.

        Returns
        -------
        bool

        """
        return self.embed_in_plane(fixed)



from numpy import array
from numpy import float64
from scipy.optimize import minimize

from compas.geometry import angle_vectors_xy
from compas.linalg import normrow
from compas.matrices import connectivity_matrix
from compas_ags.ags.core import update_primal_from_dual
from compas_ags.diagrams import ForceDiagram
from compas_ags.diagrams import FormDiagram


def compute_loadpath(form: FormDiagram, force: ForceDiagram):
    """Compute the internal work of a structure.

    Parameters
    ----------
    form : :class:`FormDiagram`
        The form diagram.
    force : :class:`ForceDiagram`
        The force diagram.

    Returns
    -------
    float
        The internal work done by the structure.

    """
    return compute_internal_work(form, force)


def compute_external_work(form: FormDiagram, force: ForceDiagram):
    """Compute the external work of a structure.

    The external work done by a structure is equal to the work done by the external
    forces. This is equal to the sum of the dot products of the force vectors and
    the vectors defined by the force application point and a fixed arbitrary point.

    Parameters
    ----------
    form : :class:`FormDiagram`
        The form diagram.
    force : :class:`ForceDiagram`
        The force diagram.

    Returns
    -------
    float
        The external work done by the structure.

    """
    vertex_index = form.vertex_index()
    xy = array(form.xy(), dtype=float64)
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    C = connectivity_matrix(edges, "csr")

    _vertex_index = force.vertex_index()
    _xy = force.xy()
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]
    _C = connectivity_matrix(_edges, "csr")

    leaves = set(form.leaves())
    external = [i for i, (u, v) in enumerate(form.edges()) if u in leaves or v in leaves]

    lengths = normrow(C.dot(xy))
    forces = normrow(_C.dot(_xy))

    return lengths[external].T.dot(forces[external])[0, 0]


def compute_internal_work(form: FormDiagram, force: ForceDiagram):
    """Compute the work done by the internal forces of a structure.

    Parameters
    ----------
    form : :class:`FormDiagram`
        The form diagram.
    force : :class:`ForceDiagram`
        The force diagram.

    Returns
    -------
    float
        The internal work done by the structure.

    """
    vertex_index = form.vertex_index()
    xy = array(form.xy(), dtype=float64)
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    C = connectivity_matrix(edges, "csr")

    _vertex_index = force.vertex_index()
    _xy = force.xy()
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]
    _C = connectivity_matrix(_edges, "csr")

    leaves = set(form.leaves())
    internal = [i for i, (u, v) in enumerate(form.edges()) if u not in leaves and v not in leaves]

    lengths = normrow(C.dot(xy))
    forces = normrow(_C.dot(_xy))

    return lengths[internal].T.dot(forces[internal])[0, 0]


def compute_internal_work_tension(form: FormDiagram, force: ForceDiagram):
    """Compute the work done by the internal tensile forces of a structure.

    Parameters
    ----------
    form : :class:`FormDiagram`
        The form diagram.
    force : :class:`ForceDiagram`
        The force diagram.

    Returns
    -------
    float
        The internal work done by the tensile forces in a structure.

    """
    vertex_index = form.vertex_index()
    xy = array(form.xy(), dtype=float64)
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    C = connectivity_matrix(edges, "csr")
    q = array(form.q(), dtype=float64).reshape((-1, 1))

    _vertex_index = force.vertex_index()
    _xy = force.xy()
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]
    _C = connectivity_matrix(_edges, "csr")

    leaves = set(form.leaves())
    internal = [i for i, (u, v) in enumerate(form.edges()) if u not in leaves and v not in leaves]
    tension = [i for i in internal if q[i, 0] > 0]

    lengths = normrow(C.dot(xy))
    forces = normrow(_C.dot(_xy))

    return lengths[tension].T.dot(forces[tension])[0, 0]


def compute_internal_work_compression(form: FormDiagram, force: ForceDiagram):
    """Compute the work done by the internal compressive forces of a structure.

    Parameters
    ----------
    form : :class:`FormDiagram`
        The form diagram.
    force : :class:`ForceDiagram`
        The force diagram.

    Returns
    -------
    float
        The internal work done by the compressive forces in a structure.

    """
    vertex_index = form.vertex_index()
    xy = array(form.xy(), dtype=float64)
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    C = connectivity_matrix(edges, "csr")
    q = array(form.q(), dtype=float64).reshape((-1, 1))

    _vertex_index = force.vertex_index()
    _xy = force.xy()
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]
    _C = connectivity_matrix(_edges, "csr")

    leaves = set(form.leaves())
    internal = [i for i, (u, v) in enumerate(form.edges()) if u not in leaves and v not in leaves]
    compression = [i for i in internal if q[i, 0] < 0]

    lengths = normrow(C.dot(xy))
    forces = normrow(_C.dot(_xy))

    return lengths[compression].T.dot(forces[compression])[0, 0]


def optimise_loadpath(form: FormDiagram, force, algo="COBYLA"):
    """Optimise the loadpath using the parameters of the force domain. The parameters
    of the force domain are the coordinates of the vertices of the force diagram.

    Parameters
    ----------
    form : :class:`FormDiagram`
        The form diagram.
    force : :class:`ForceDiagram`
        The force diagram.
    algo : {'COBYLA', L-BFGS-B', 'SLSQ', 'MMA', 'GMMA'}, optional
        The optimisation algorithm.

    Returns
    -------
    form: :class:`FormDiagram`
        The optimised form diagram.
    force: :class:`ForceDiagram`
        The optimised force diagram.

    Notes
    -----
    In many cases, the number of paramters of the force domain involved in generating
    new solutions in the form domain is smaller than when using the elements of the
    form diagram directly.

    For example, the loadpath of a bridge with a compression arch can be optimised
    using only the x-coordinates of the vertices of the force diagram corresponding
    to internal spaces formed by the segments of the arch, the corresponding deck
    elements, and the hangers connecting them. Any solution generated by these parameters
    will be in equilibrium and automatically have a horizontal bridge deck.

    Although the *BRG* algorithm is the preferred choice, since it is (should be)
    tailored to the problem of optimising loadpaths using the domain of the force
    diagram, it does not have a stable and/or efficient implementation.
    The main problem is the generation of the form diagram, based on a given force
    diagram. For example, when edge forces flip from tension to compression, and
    vice versa, parallelisation is no longer effective.

    """
    vertex_index = form.vertex_index()
    edge_index = form.edge_index()
    i_j = {vertex_index[vertex]: [vertex_index[nbr] for nbr in form.vertex_neighbors(vertex)] for vertex in form.vertices()}
    ij_e = {(vertex_index[u], vertex_index[v]): edge_index[u, v] for u, v in edge_index}
    ij_e.update({(vertex_index[v], vertex_index[u]): edge_index[u, v] for u, v in edge_index})

    xy = array(form.xy(), dtype=float64)
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    C = connectivity_matrix(edges, "csr")

    leaves = [vertex_index[key] for key in form.leaves()]
    fixed = [vertex_index[key] for key in form.fixed()]
    free = list(set(range(form.number_of_vertices())) - set(fixed) - set(leaves))
    internal = [i for i, (u, v) in enumerate(form.edges()) if vertex_index[u] not in leaves and vertex_index[v] not in leaves]

    _vertex_index = force.vertex_index()
    _edge_index = force.edge_index(form)
    _edge_index.update({(v, u): _edge_index[u, v] for u, v in _edge_index})

    _xy = array(force.xy(), dtype=float64)
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]
    _C = connectivity_matrix(_edges, "csr")

    _free = [key for key, attr in force.vertices(True) if attr["is_param"]]
    _free = [_vertex_index[key] for key in _free]

    def objfunc(_x):
        _xy[_free, 0] = _x

        update_primal_from_dual(xy, _xy, free, i_j, ij_e, _C, leaves=leaves)

        length = normrow(C.dot(xy))
        force = normrow(_C.dot(_xy))
        lp = length[internal].T.dot(force[internal])[0, 0]

        print(lp)
        return lp

    x0 = _xy[_free, 0]

    result = minimize(objfunc, x0, method=algo, tol=1e-12, options={"maxiter": 1000})  # noqa: F841

    uv = C.dot(xy)
    _uv = _C.dot(_xy)
    angles = [angle_vectors_xy(a, b) for a, b in zip(uv, _uv)]
    lengths = normrow(uv)
    forces = normrow(_uv)
    q = forces / lengths

    for vertex, attr in form.vertices(True):
        index = vertex_index[vertex]
        attr["x"] = xy[index, 0]
        attr["y"] = xy[index, 1]

    for edge, attr in form.edges(True):
        index = edge_index[edge]
        attr["l"] = lengths[index, 0]
        attr["a"] = angles[index]
        if (angles[index] - 3.14159) ** 2 < 0.25 * 3.14159:
            attr["f"] = -forces[index, 0]
            attr["q"] = -q[index, 0]
        else:
            attr["f"] = forces[index, 0]
            attr["q"] = q[index, 0]

    for vertex, attr in force.vertices(True):
        index = _vertex_index[vertex]
        attr["x"] = _xy[index, 0]
        attr["y"] = _xy[index, 1]

    for edge, attr in force.edges(True):
        index = _edge_index[edge]
        attr["a"] = angles[index]
        attr["l"] = forces[index, 0]

    return form, force



from compas.datastructures import Mesh


class Diagram(Mesh):
    """Basic mesh-based data structure for diagrams in AGS.

    Attributes
    ----------
    dual : :class:`compas_ags.diagrams.Diagram`
        The dual diagram of this diagram.

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._dual = None

    @property
    def dual(self):
        return self._dual

    @dual.setter
    def dual(self, dual):
        self._dual = dual

    def vertex_index(self):
        return {vertex: index for index, vertex in enumerate(self.vertices())}

    def index_vertex(self):
        return {index: vertex for index, vertex in enumerate(self.vertices())}

    def edge_index(self):
        return {edge: index for index, edge in enumerate(self.edges())}

    def index_edge(self):
        return {index: edge for index, edge in enumerate(self.edges())}



import sys

import numpy.typing as npt
from numpy import array
from numpy import asarray
from numpy import atleast_2d
from numpy import delete
from numpy import diag
from numpy import eye
from numpy import float64
from numpy import hstack
from numpy import vstack
from numpy import zeros
from numpy.linalg import cond
from numpy.linalg import inv
from numpy.linalg import matrix_rank
from scipy.linalg import lstsq
from scipy.linalg import solve
from scipy.sparse import spmatrix

from compas.geometry import Line
from compas.geometry import midpoint_point_point_xy
from compas.geometry import project_point_line_xy
from compas.linalg import normalizerow
from compas.linalg import normrow
from compas.linalg import solve_with_known
from compas.matrices import connectivity_matrix
from compas.matrices import equilibrium_matrix
from compas.matrices import laplacian_matrix
from compas_ags.exceptions import SolutionError

EPS = 1 / sys.float_info.epsilon


def rref_sympy(A, tol=None):
    r"""Reduced row-echelon form of matrix A.

    Parameters
    ----------
    A : array-like
        Matrix A represented as an array or list.
    tol : float
        Tolerance.

    Returns
    -------
    array
        RREF of A.

    Notes
    -----
    A matrix is in reduced row-echelon form after Gauss-Jordan elimination, the
    result is independent of the method/algorithm used.

    Examples
    --------
    >>> A = [[1, 0, 1, 3], [2, 3, 4, 7], [-1, -3, -3, -4]]
    >>> n = rref_sympy(A)
    >>> array(n)
    array([[1, 0, 1.00000000000000, 3.00000000000000],
           [0, 1, 0.666666666666667, 0.333333333333333],
           [0, 0, 0, 0]], dtype=object)

    """
    import sympy

    A = atleast_2d(asarray(A, dtype=float))
    return sympy.Matrix(A).rref()[0].tolist()


def update_q_from_qind(E: spmatrix, q: npt.NDArray, dep: list[int], ind: list[int]) -> None:
    """Update the full set of force densities using the values of the independent edges.

    Parameters
    ----------
    E : sparse csr matrix
        The equilibrium matrix.
    q : array
        The force densities of the edges.
    dep : list
        The indices of the dependent edges.
    ind : list
        The indices of the independent edges.

    Returns
    -------
    None
        The force densities are modified in-place.

    """
    m = E.shape[0] - len(dep)
    qi = q[ind]
    Ei = E[:, ind]
    Ed = E[:, dep]
    if m > 0:
        Edt = Ed.transpose()
        A = Edt.dot(Ed).toarray()
        b = Edt.dot(Ei).dot(qi)
    else:
        A = Ed.toarray()
        b = Ei.dot(qi)
    if cond(A) > EPS:
        res = lstsq(-A, b)
        qd = res[0]
    else:
        qd = solve(-A, b)
    q[dep] = qd


def update_primal_from_dual(
    xy: npt.ArrayLike,
    _xy: npt.ArrayLike,
    free: list[int],
    i_nbrs: list[list[int]],
    ij_e: dict[tuple[int, int], int],
    _C: spmatrix,
    line_constraints: list = None,
    target_lengths: list[float] = [],
    target_vectors: list[list[float]] = [],
    leaves: list[int] = [],
    kmax: int = 100,
) -> None:
    r"""Update the coordinates of the primal diagram using the coordinates of the corresponding dual diagram.
    This function apply to both sides, i.e. it can be used to update the form diagram from the geometry of the force
    diagram or to update the force diagram from the geometry of the force diagram.

    Parameters
    ----------
    xy : array-like
        XY coordinates of the vertices of the primal diagram.
    _xy : array-like
        XY coordinates of the vertices of the dual diagram.
    free : list
        List of free vertices.
    i_nbrs : list of list of int
        Vertex neighbours per vertex.
    ij_e : dict
        Edge index for every vertex pair.
    _C : sparse matrix in csr format
        The connectivity matrix of the force diagram.
    line_constraints : list, optional
        Line constraints applied to the free nodes.
        Default is an ``None`` in which case no line constraints are considered.
    target_lengths : list, optional
        Target lengths / target forces of the edges.
        Default is an empty list, which considers that no target lengths are considered.
    target_vectors : list, optional
        Target vectors of the edges.
        Default is an empty list, which considers that no target vectors are considered.
    leaves : list, optional
        The leaves of the primal diagram.
        Default is an empty list, which considers that no leaves are considered.
    kmax : int, optional
        Maximum number of iterations.
        Default is ``100``.

    Returns
    -------
    None
        The vertex coordinates are modified in-place.

    Notes
    -----
    This function should be used to update the form diagram after modifying the
    geometry of the force diagram. Or to update the force diagram geometrically to
    become reciprocal to the form diagram. The objective is to compute new locations
    for the vertices of the primal diagram such that the corrsponding lines of the
    primal and dual diagram are parallel while any geometric constraints imposed on
    the primal diagram are satisfied. Note that form, and force can assume position
    of primal and dual interchagable.

    The location of each vertex of the primal diagram is computed as the intersection
    of the lines connected to it. Each of the connected lines is based at the connected
    neighbouring vertex and taken parallel to the corresponding line in the dual
    diagram.

    For a point :math:`\mathbf{p}`, which is the least-squares intersection of *K*
    lines, with every line *j* defined by a point :math:`\mathbf{a}_{j}` on the line
    and a direction vector :math:`\mathbf{n}_{j}`, we can write

    .. math::

        \mathbf{R} \mathbf{p} = \mathbf{q}

    with

    .. math::

        \mathbf{R} = \displaystyle\sum_{j=1}^{K}(\mathbf{I} - \mathbf{n}_{j}\mathbf{n}_{j}^{T})
        \quad,\quad
        \mathbf{q} = \displaystyle\sum_{j=1}^{K}(\mathbf{I} - \mathbf{n}_{j}\mathbf{n}_{j}^{T})\mathbf{a}_{j}

    This system of linear equations can be solved using the normal equations

    .. math::

        \mathbf{p} = (\mathbf{R}^{T}\mathbf{R})^{-1}\mathbf{R}^{T}\mathbf{q}

    """
    _uv = _C.dot(_xy)
    _t = normalizerow(_uv)
    I = eye(2, dtype=float64)  # noqa: E741
    xy0 = array(xy, copy=True)
    A = zeros((2 * len(free), 2 * len(free)), dtype=float64)
    b = zeros((2 * len(free), 1), dtype=float64)

    # update the free vertices
    for k in range(kmax):
        row = 0

        # in order for the two diagrams to have parallel corresponding edges,
        # each free vertex location of the primal diagram is computed as the intersection
        # of the connected lines. each of these lines is based at the corresponding
        # connected neighbouring vertex and taken parallel to the corresponding
        # edge in the dual diagram.
        # the intersection is the point that minimises the distance to all connected
        # lines.
        for count in range(len(free)):
            i = free[count]
            R = zeros((2, 2), dtype=float64)
            q = zeros((2, 1), dtype=float64)

            # add line constraints based on connected edges
            for j in i_nbrs[i]:
                if j in leaves:
                    continue

                n = _t[ij_e[(i, j)], None]  # the direction of the line (the line parallel to the corresponding line in the force diagram)
                _l = _uv[ij_e[(i, j)], None]

                if normrow(_l)[0, 0] < 0.001:
                    continue

                if target_lengths:
                    if target_lengths[ij_e[(i, j)]] == 0.0:
                        continue

                if target_vectors:
                    if target_vectors[ij_e[(i, j)]]:
                        n = array(target_vectors[ij_e[(i, j)]]).reshape(1, -1)

                r = I - n.T.dot(n)  # projection into the orthogonal space of the direction vector
                a = xy[j, None]  # a point on the line (the neighbour of the vertex)
                R += r
                q += r.dot(a.T)

            if line_constraints:
                line: Line = line_constraints[count]
                if line:
                    n_ = array(line.direction[:2]).reshape(1, 2)
                    r = I - n_.T.dot(n_)
                    pt = array(line.start[:2]).reshape(1, 2)
                    R += r
                    q += r.dot(pt.T)

            A[row : row + 2, row : row + 2] = R
            b[row : row + 2] = q
            row += 2

            # p = solve(R.T.dot(R), R.T.dot(q))
            # xy[i] = p.reshape((-1, 2), order='C')

        res = lstsq(A, b)
        xy_lstsq = res[0].reshape((-1, 2), order="C")
        xy[free] = xy_lstsq

    # reconnect leaves
    for i in leaves:
        j = i_nbrs[i][0]
        xy[i] = xy[j] + xy0[i] - xy0[j]


def parallelise_edges(
    xy,
    edges,
    i_nbrs,
    ij_e,
    target_vectors,
    target_lengths,
    fixed=None,
    line_constraints=None,
    kmax=100,
    callback=None,
):
    """Parallelise the edges of a mesh to given target vectors.

    Parameters
    ----------
    xy : list
        The XY coordinates of the vertices of the edges.
    edges : list
        The edges as pairs of indices in ``xy``.
    i_nbrs : dict
        A list of neighbours per vertex.
    ij_e : dict
        An edge index per vertex pair.
    target_vectors : list
        A list with an entry for each edge representing the target vector or ``None``.
    target_lengths : list
        A list with an entry for each edge representing the target length or ``None``.
    fixed : list, optional
        The fixed nodes of the mesh.
        Default is ``None``.
    line_constraints : list, optional
        Line constraints applied to the free nodes.
        Default is an ``None`` in which case no line constraints are considered.
    kmax : int, optional
        Maximum number of iterations.
        Default is ``100``.
    callback : callable, optional
        A user-defined callback function to be executed after every iteration.
        Default is ``None``.

    Returns
    -------
    None

    Notes
    -----
    This implementation is based on the function ``compas_tna.equilibrium.parallelisation.parallelise_edges``.

    """
    if callback:
        if not callable(callback):
            raise Exception("The provided callback is not callable.")

    n = len(xy)

    for k in range(kmax):
        xy0 = [[x, y] for x, y in xy]
        uv = [[xy[j][0] - xy[i][0], xy[j][1] - xy[i][1]] for i, j in edges]
        lengths = [(dx**2 + dy**2) ** 0.5 for dx, dy in uv]

        for j in range(n):
            if j in fixed:
                continue

            nbrs = i_nbrs[j]
            x, y = 0.0, 0.0

            len_nbrs = 0
            for i in nbrs:
                if (i, j) in ij_e:
                    e = ij_e[(i, j)]
                    u, v = i, j
                    signe = +1.0
                else:
                    e = ij_e[(j, i)]
                    u, v = j, i
                    signe = -1.0

                if target_lengths[e] is not None:  # edges with constraint on length ...
                    lij = target_lengths[e]
                    if target_vectors[e]:  # edges with constraint on length + orientation
                        tx, ty = target_vectors[e]
                    else:  # edges with constraint on length only
                        if lengths[e] == 0.0:
                            tx = ty = 0.0
                        else:
                            tx = (xy0[v][0] - xy0[u][0]) / lengths[e]
                            ty = (xy0[v][1] - xy0[u][1]) / lengths[e]  # check if xy0 is indeed better than xy
                else:
                    if target_vectors[e]:  # edges with constraint on orientation only
                        tx, ty = target_vectors[e]
                        lij = lengths[e]
                    else:
                        continue  # edges to discard

                ax, ay = xy0[i]
                x += ax + signe * lij * tx
                y += ay + signe * lij * ty
                len_nbrs += 1

            if len_nbrs > 0:
                xy[j][0] = x / len_nbrs
                xy[j][1] = y / len_nbrs

                # check if line constraints are applied and project result
                if line_constraints:
                    line = line_constraints[j]
                    if line:
                        pt_proj = project_point_line_xy(xy[j], line)
                        xy[j][0] = pt_proj[0]
                        xy[j][1] = pt_proj[1]

        for i, j in ij_e:
            e = ij_e[(i, j)]

            if lengths[e] == 0.0 or target_lengths[e] == 0.0:
                c = midpoint_point_point_xy(xy[i], xy[j])
                xy[i][:] = c[:][:2]
                xy[j][:] = c[:][:2]

        if callback:
            callback(k, xy, edges)


def get_jacobian_and_residual(form, force, _X_goal, constraints=None):
    r"""Compute the Jacobian matrix and residual.

    Computes the residual and the Jacobian matrix :math:`\partial \mathbf{X}^* / \partial \mathbf{X}`
    where :math:`\mathbf{X}` contains the form diagram coordinates in *Fortran* order
    (first all :math:`\mathbf{x}`-coordinates, then all :math:`\mathbf{y}`-coordinates) and :math:`\mathbf{X}^*` contains the
    force diagram coordinates in *Fortran* order (first all :math:`\mathbf{x}^*`-coordinates,
    then all :math:`\mathbf{y}^*`-coordinates). Jacobian and residual have the rows corresponding to the
    force diagrarm anchor vertex removed.

    Parameters
    ----------
    form: :class:`FormDiagram`
        The form diagram to update.
    force: :class:`ForceDiagram`
        The force diagram on which the update is based.
    _X_goal: array [2*n]
        Contains the target force diagram coordinates (:math:`\mathbf{X}^*`) in *Fortran* order
        (first all :math:`\mathbf{x}^*`-coordinates, then all :math:`\mathbf{y}^*`-coordinates).
    constraints: :class:`ConstraintsCollection`, optional
        A collection of form diagram constraints.
        The default is ``None``, in which case no constraints are considered.

    Returns
    -------
    red_jacobian, red_r: tuple of arrays
        Jacobian matrix and residual vector as arrays.
        The rows corresponding to the anchor of the force diagram are removed

    References
    ----------
    .. [1] Alic, V. and Åkesson, D., 2017. Bi-directional algebraic graphic statics. Computer-Aided Design, 93, pp.26-37.

    """

    jacobian = compute_jacobian(form, force)

    _vcount = force.number_of_vertices()
    _k_i = force.key_index()
    _known = _k_i[force.anchor()]
    _bc = [_known, _vcount + _known]
    _X_iteration = array(force.vertices_attribute("x") + force.vertices_attribute("y")).reshape(-1, 1)
    r = _X_iteration - _X_goal

    if constraints:
        (cj, cr) = constraints.compute_constraints()
        jacobian = vstack((jacobian, cj))
        r = vstack((r, cr))

    # Check rank of augmented matrix
    rank_jac = matrix_rank(jacobian)
    rank_aug = matrix_rank(hstack([jacobian, r]))

    if rank_jac < rank_aug:
        raise SolutionError("ERROR: Rank Augmented > Rank Jacobian")

    # Remove rows due to anchored vertex in the force diagram
    red_r = delete(r, _bc, axis=0)
    red_jacobian = delete(jacobian, _bc, axis=0)

    return red_jacobian, red_r


def compute_jacobian(form, force):
    r"""Compute the Jacobian matrix.

    The actual computation of the Jacobian matrix :math:`\partial \mathbf{X}^* / \partial \mathbf{X}`
    where :math:`\mathbf{X}` contains the form diagram coordinates in *Fortran* order
    (first all :math:`\mathbf{x}`-coordinates, then all :math:`\mathbf{y}`-coordinates) and :math:`\mathbf{X}^*` contains the
    force diagram coordinates in *Fortran* order (first all :math:`\mathbf{x}^*`-coordinates,
    then all :math:`\mathbf{y}^*`-coordinates).

    Parameters
    ----------
    form: :class:`FormDiagram`
        The form diagram.
    force: :class:`ForceDiagram`
        The force diagram.

    Returns
    -------
    jacobian
        Jacobian matrix (2 * _vcount, 2 * vcount)

    References
    ----------
    .. [1] Alic, V. and Åkesson, D., 2017. Bi-directional algebraic graphic statics. Computer-Aided Design, 93, pp.26-37.

    """
    # --------------------------------------------------------------------------
    # form diagram
    # --------------------------------------------------------------------------
    vcount = form.number_of_vertices()
    k_i = form.key_index()
    leaves = [k_i[key] for key in form.leaves()]
    free = list(set(range(form.number_of_vertices())) - set(leaves))
    vicount = len(free)
    edges = [(k_i[u], k_i[v]) for u, v in form.edges()]
    xy = array(form.xy(), dtype=float64).reshape((-1, 2))
    ecount = len(edges)
    C = connectivity_matrix(edges, "array")
    E = equilibrium_matrix(C, xy, free, "array")
    uv = C.dot(xy)
    u = uv[:, 0].reshape(-1, 1)
    v = uv[:, 1].reshape(-1, 1)
    Ct = C.transpose()
    Cti = array(Ct[free, :])

    q = array(form.q(), dtype=float64).reshape((-1, 1))
    Q = diag(q.flatten())  # TODO: Explore sparse (diags)

    independent_edges = [(k_i[u], k_i[v]) for (u, v) in list(form.edges_where({"is_ind": True}))]
    independent_edges_idx = [edges.index(i) for i in independent_edges]
    dependent_edges_idx = list(set(range(ecount)) - set(independent_edges_idx))

    Ed = E[:, dependent_edges_idx]
    Eid = E[:, independent_edges_idx]
    qid = q[independent_edges_idx]
    EdInv = inv(array(Ed))  # TODO: Explore sparse (spinv)

    # --------------------------------------------------------------------------
    # force diagram
    # --------------------------------------------------------------------------
    _vertex_index = force.vertex_index()
    _vcount = force.number_of_vertices()
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]
    _L = laplacian_matrix(_edges, normalize=False, rtype="array")
    _C = connectivity_matrix(_edges, "array")
    _Ct = _C.transpose()
    _Ct = array(_Ct)
    _known = [_vertex_index[force.anchor()]]

    # --------------------------------------------------------------------------
    # Jacobian
    # --------------------------------------------------------------------------
    jacobian = zeros((_vcount * 2, vcount * 2))
    for j in range(2):  # Loop for x and y
        idx = list(range(j * vicount, (j + 1) * vicount))
        for i in range(vcount):
            dXdxi = diag(Ct[i, :])
            dxdxi = Ct[i, :].reshape(-1, 1)

            dEdXi = zeros((vicount * 2, ecount))
            dEdXi[idx, :] = Cti.dot(dXdxi)

            dEdXi_d = dEdXi[:, dependent_edges_idx]
            dEdXi_id = dEdXi[:, independent_edges_idx]

            dEdXiInv = -EdInv.dot(dEdXi_d.dot(EdInv))

            dqdXi_d = -dEdXiInv.dot(Eid.dot(qid)) - EdInv.dot(dEdXi_id.dot(qid))
            dqdXi = zeros((ecount, 1))
            dqdXi[dependent_edges_idx] = dqdXi_d
            dqdXi[independent_edges_idx] = 0
            dQdXi = diag(dqdXi[:, 0])

            d_XdXiTop = zeros((_L.shape[0]))
            d_XdXiBot = zeros((_L.shape[0]))

            if j == 0:
                d_XdXiTop = solve_with_known(
                    _L,
                    (_Ct.dot(dQdXi.dot(u) + Q.dot(dxdxi))).flatten(),
                    d_XdXiTop,
                    _known,
                )
                d_XdXiBot = solve_with_known(_L, (_Ct.dot(dQdXi.dot(v))).flatten(), d_XdXiBot, _known)
            elif j == 1:
                d_XdXiTop = solve_with_known(_L, (_Ct.dot(dQdXi.dot(u))).flatten(), d_XdXiTop, _known)
                d_XdXiBot = solve_with_known(
                    _L,
                    (_Ct.dot(dQdXi.dot(v) + Q.dot(dxdxi))).flatten(),
                    d_XdXiBot,
                    _known,
                )

            d_XdXi = hstack((d_XdXiTop, d_XdXiBot))
            jacobian[:, i + j * vcount] = d_XdXi
    return jacobian



from typing import Annotated
from typing import Callable
from typing import Literal

import numpy.typing as npt
from numpy import array
from numpy import delete
from numpy import float64
from numpy import vstack
from numpy.linalg import lstsq
from numpy.linalg import norm
from scipy.sparse import diags

from compas.geometry import angle_vectors_xy
from compas.linalg import dof
from compas.linalg import nonpivots
from compas.linalg import normrow
from compas.linalg import nullspace as matrix_nullspace
from compas.linalg import spsolve_with_known
from compas.matrices import connectivity_matrix
from compas.matrices import equilibrium_matrix
from compas_ags.ags.constraints import ConstraintsCollection
from compas_ags.ags.core import compute_jacobian
from compas_ags.ags.core import get_jacobian_and_residual
from compas_ags.ags.core import parallelise_edges
from compas_ags.ags.core import update_primal_from_dual
from compas_ags.ags.core import update_q_from_qind
from compas_ags.diagrams import ForceDiagram
from compas_ags.diagrams import FormDiagram
from compas_ags.exceptions import SolutionError

from .core import rref_sympy

# ==============================================================================
# analysis form diagram
# ==============================================================================


def form_identify_dof(form: FormDiagram) -> tuple[int, int, list[int]]:
    r"""Identify the DOF of a form diagram.

    Parameters
    ----------
    form: :class:`FormDiagram`
        The form diagram.

    Returns
    -------
    k : int
        Dimension of the null space (nullity) of the equilibrium matrix.
        Number of independent states of self-stress.
    m : int
        Size of the left null space of the equilibrium matrix.
        Number of (infenitesimal) mechanisms.
    ind : list
        Indices of the independent edges.

    Notes
    -----
    The equilibrium matrix of the form diagram is

    .. math::

        \mathbf{E}
        =
        \begin{bmatrix}
        \mathbf{C}_{i}^{t}\mathbf{U} \\
        \mathbf{C}_{i}^{t}\mathbf{V}
        \end{bmatrix}

    If ``k == 0`` and ``m == 0``, the system described by the equilibrium matrix is statically determined.
    If ``k > 0`` and ``m == 0``, the system is statically indetermined with `k` idependent states of stress.
    If ``k == 0`` and ``m > 0``, the system is unstable, with `m` independent mechanisms.

    The dimension of a vector space (such as the null space) is the number of
    vectors of a basis of that vector space. A set of vectors forms a basis of a
    vector space if they are linearly independent vectors and every vector of the
    space is a linear combination of this set.

    """
    vertex_index = form.vertex_index()

    xy = form.vertices_attributes("xy")
    fixed = [vertex_index[vertex] for vertex in form.fixed()]
    free = list(set(range(form.number_of_vertices())) - set(fixed))
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    C = connectivity_matrix(edges)
    E = equilibrium_matrix(C, xy, free)

    k, m = dof(E)
    ind = nonpivots(rref_sympy(E))

    return int(k), int(m), [edges[i] for i in ind]


def form_count_dof(form: FormDiagram) -> tuple[int, int]:
    r"""Count the number of degrees of freedom of a form diagram.

    Parameters
    ----------
    form: :class:`FormDiagram`
        The form diagram.

    Returns
    -------
    k : int
        Dimension of the null space (*nullity*) of the equilibrium matrix of the
        form diagram.
    m : int
        Dimension of the left null space of the equilibrium matrix of the form
        diagram.

    Notes
    -----
    The equilibrium matrix of the form diagram is

    .. math::

        \mathbf{E}
        =
        \begin{bmatrix}
        \mathbf{C}_{i}^{t}\mathbf{U} \\
        \mathbf{C}_{i}^{t}\mathbf{V}
        \end{bmatrix}

    """
    vertex_index = form.vertex_index()

    xy = form.vertices_attributes("xy")
    fixed = [vertex_index[vertex] for vertex in form.leaves()]
    free = list(set(range(form.number_of_vertices())) - set(fixed))
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    C = connectivity_matrix(edges)
    E = equilibrium_matrix(C, xy, free)

    k, m = dof(E)

    return int(k), int(m)


def form_compute_nullspace(
    form: FormDiagram,
    force: ForceDiagram,
    constraints: ConstraintsCollection = None,
) -> list[Annotated[npt.NDArray, Literal["N", 2]]]:
    r"""Compute the nullspaces of a form diagram assuming a set of constraints.

    It returns a list with the displacements that apply to the form diagram representing
    how the latter can be moved in the plan without modifying the force diagram and
    considering the set of constraints applied. It corresponds to the nullspace of the
    Jacobian :math:`\partial \mathbf{X}^* / \partial \mathbf{X}` matrix as computed in [1].

    Parameters
    ----------
    form : :class:`FormDiagram`
        The form diagram.
    force : :class:`ForceDiagram`
        The force diagram.
    constraints: :class:`ConstraintsCollection`, optional
        A collection of form diagram constraints.
        The default is ``None``, in which case no constraints are considered.

    Returns
    -------
    nullspaces [list of arrays (vcount x 2)]
        The null displacement fields applied to the form diagram considering applied constraints.

    Notes
    -----
    Among the nullspaces, the unrestrained rigid-body displacements available are always
    included and other unrestrained displacement fields that preserve the orientation
    of all edges keeping reciprocity between form and force diagram. The displacement fields
    are unit vectors and can be rescaled to visialisation purposes.

    References
    ----------
    .. [1] Alic, V. and Åkesson, D., 2017. Bi-directional algebraic graphic statics. Computer-Aided Design, 93, pp.26-37.

    """
    jacobian = compute_jacobian(form, force)  # Jacobian matrix of size (2 _vcount, 2 vcount)
    if constraints:
        (cj, _) = constraints.compute_constraints()
        jacobian = vstack((jacobian, cj))  # Add rows to the Jacobian matrix representing constraints

    # Remove the rows of the jacobian to account for the anchored vertex in the force diagram (influence x and y directions)
    _vcount = force.number_of_vertices()
    _k_i = force.vertex_index()
    _anchor = _k_i[force.anchor()]
    _anchor_xy = [_anchor, _vcount + _anchor]

    reduced_jacobian = delete(jacobian, _anchor_xy, axis=0)

    nullstates = matrix_nullspace(reduced_jacobian).T  # unit vectors representing the possible nullstates

    nullspaces = []
    for nullstate in nullstates:  # reshaping the nullstates a list of (vcount x 2) arrays
        xy = nullstate.reshape((2, -1)).T
        nullspaces.append(xy)

    return nullspaces


# ==============================================================================
# update form diagram
# ==============================================================================


def form_update_q_from_qind(form: FormDiagram) -> FormDiagram:
    """Update the force densities of the dependent edges of a form diagram using
    the values of the independent ones.

    Parameters
    ----------
    form: :class:`FormDiagram`
        The form diagram.

    Returns
    -------
    form: :class:`FormDiagram`
        The updated form diagram with force densities stored as attributes of the edges.

    Examples
    --------
    >>>
    """
    vertex_index = form.vertex_index()
    edge_index = form.edge_index()

    vcount = form.number_of_vertices()
    ecount = form.number_of_edges()
    fixed = form.leaves()
    fixed = [vertex_index[vertex] for vertex in fixed]
    free = list(set(range(vcount)) - set(fixed))
    ind = [edge_index[edge] for edge in form.ind()]
    dep = list(set(range(ecount)) - set(ind))
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    xy = array(form.xy(), dtype=float64).reshape((-1, 2))
    q = array(form.q(), dtype=float64).reshape((-1, 1))
    C = connectivity_matrix(edges, "csr")
    E = equilibrium_matrix(C, xy, free, "csr")

    update_q_from_qind(E, q, dep, ind)

    uv = C.dot(xy)
    lengths = normrow(uv)
    forces = q * lengths

    for edge in form.edges():
        index = edge_index[edge]
        form.edge_attributes(edge, ["q", "f", "l"], [q[index, 0], forces[index, 0], lengths[index, 0]])

    return form


def form_update_from_force(form: FormDiagram, force: ForceDiagram, kmax: int = 100) -> tuple[FormDiagram, ForceDiagram]:
    r"""Update the form diagram after a modification of the force diagram.

    Parameters
    ----------
    form: :class:`FormDiagram`
        The form diagram to update.
    force : :class:`ForceDiagram`
        The force diagram on which the update is based.
    kmax: int, optional
        Maximum number of least-square iterations for solving the duality form-force.
        The default value is ``20``.

    Returns
    -------
    form: :class:`FormDiagram`
        The form diagram with updated force densities.
    force: :class:`ForceDiagram`
        The updated force diagram.

    Notes
    -----
    Compute the geometry of the form diagram from the geometry of the form diagram
    and some constraints (location of fixed points).
    Since both diagrams are reciprocal, the coordinates of each vertex of the form
    diagram can be expressed as the intersection of three or more lines parallel
    to the corresponding edges of the force diagram.
    Essentially, this boils down to solving the following problem:

    .. math::

        \mathbf{A}\mathbf{x} = \mathbf{b}

    with :math:`\mathbf{A}` the coefficients of the x and y-coordinate of the  ,
    :math:`\mathbf{x}` the coordinates of the vertices of the form diagram,
    in *Fortran* order (first all x-coordinates, then all y-coordinates),
    and  :math:`\mathbf{b}` ....

    """
    # --------------------------------------------------------------------------
    # form diagram
    # --------------------------------------------------------------------------
    vertex_index = form.vertex_index()
    edge_index = form.edge_index()
    i_j = {index: [vertex_index[nbr] for nbr in form.vertex_neighbors(vertex)] for index, vertex in enumerate(form.vertices())}
    ij_e = {(vertex_index[u], vertex_index[v]): edge_index[u, v] for u, v in edge_index}
    ij_e.update({(vertex_index[v], vertex_index[u]): edge_index[u, v] for u, v in edge_index})

    xy = array(form.xy(), dtype=float64)
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    C = connectivity_matrix(edges, "csr")
    # --------------------------------------------------------------------------
    # constraints
    # --------------------------------------------------------------------------
    leaves = [vertex_index[vertex] for vertex in form.leaves()]
    fixed = [vertex_index[vertex] for vertex in form.fixed()]
    free = list(set(range(form.number_of_vertices())) - set(fixed) - set(leaves))
    line_constraints_all = form.vertices_attribute("line_constraint")
    line_constraints = [line_constraints_all[i] for i in free]
    target_vectors = form.edges_attribute("target_vector")
    # --------------------------------------------------------------------------
    # force diagram
    # --------------------------------------------------------------------------
    _vertex_index = force.vertex_index()
    _edge_index = force.edge_index(form)
    _edge_index.update({(v, u): _edge_index[u, v] for u, v in _edge_index})

    _xy = array(force.xy(), dtype=float64)
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]
    _C = connectivity_matrix(_edges, "csr")
    # --------------------------------------------------------------------------
    # compute the coordinates of thet *free* vertices
    # as a function of the fixed vertices and the previous coordinates of the *free* vertices
    # re-add the leaves and leaf-edges
    # --------------------------------------------------------------------------
    update_primal_from_dual(
        xy,
        _xy,
        free,
        i_j,
        ij_e,
        _C,
        line_constraints=line_constraints,
        target_vectors=target_vectors,
        leaves=leaves,
        kmax=kmax,
    )
    # --------------------------------------------------------------------------
    # update
    # --------------------------------------------------------------------------
    uv = C.dot(xy)
    _uv = _C.dot(_xy)
    angles = [angle_vectors_xy(a, b, deg=True) for a, b in zip(uv, _uv)]
    lengths = normrow(uv)
    forces = normrow(_uv)
    q = forces / lengths
    # --------------------------------------------------------------------------
    # update form diagram
    # --------------------------------------------------------------------------
    for vertex, attr in form.vertices(True):
        index = vertex_index[vertex]
        attr["x"] = xy[index, 0]
        attr["y"] = xy[index, 1]
    for edge, attr in form.edges(True):
        index = edge_index[edge]
        attr["l"] = lengths[index, 0]
        attr["a"] = angles[index]
        if angles[index] < 90:
            attr["f"] = forces[index, 0]
            attr["q"] = q[index, 0]
        else:
            attr["f"] = -forces[index, 0]
            attr["q"] = -q[index, 0]
    # --------------------------------------------------------------------------
    # update force diagram
    # --------------------------------------------------------------------------
    for edge, attr in force.edges(True):
        index = _edge_index[edge]
        attr["a"] = angles[index]
        attr["l"] = forces[index, 0]

    return form, force


def form_update_from_force_newton(
    form: FormDiagram,
    force: ForceDiagram,
    constraints: ConstraintsCollection = None,
    tol: float = 1e-10,
    max_iter: int = 20,
) -> FormDiagram:
    r"""Update the form diagram after a modification of the force diagram.

    Compute the geometry of the form diagram from the geometry of the force diagram
    and some constraints, including non-linear.
    The form diagram is computed by formulating the root-finding approach presented
    in bi-dir AGS [1]_. Newton's method is used to solve for the form diagram.

    The algorithm fails if it is over-constrained or if the initial guess is too far
    from any root.

    Parameters
    ----------
    form: :class:`FormDiagram`
        The form diagram to update.
    force: :class:`ForceDiagram`
        The force diagram containing the vertex modification desired.
    constraints: :class:`ConstraintsCollection`, optional
        A collection of form diagram constraints.
        The default is ``None``, in which case no constraints are considered.
    tol: float, optional
        Stopping criteria tolerance.
        The default value is ``1e-10``.
    max_iter: int, optional
        Maximum number of iterations before stop Newton Method.
        The default value is ``20``.

    Returns
    -------
    form: :class:`FormDiagram`
        The updated form diagram.

    References
    ----------
    .. [1] Alic, V. and Åkesson, D., 2017. Bi-directional algebraic graphic statics. Computer-Aided Design, 93, pp.26-37.

    """
    X = array(form.vertices_attribute("x") + form.vertices_attribute("y")).reshape(-1, 1)
    _X_goal = array(force.vertices_attribute("x") + force.vertices_attribute("y")).reshape(-1, 1)

    vcount = form.number_of_vertices()
    index_vertex = form.index_vertex()

    # Begin Newton
    diff = 100
    n_iter = 1
    while diff > tol:
        # Update force diagram based on form at each iteration
        form_update_q_from_qind(form)
        force_update_from_form(force, form)

        # Get jacobian maxtrix and residual vector considering constraints
        red_jacobian, red_r = get_jacobian_and_residual(form, force, _X_goal, constraints)

        # Do the least squares solution
        dx = lstsq(red_jacobian, -red_r)[0]

        X = X + dx

        # Update form diagram at end of iteration
        for i in range(vcount):
            vertex = index_vertex[i]
            form.vertex_attribute(vertex, "x", X[i].item())
            form.vertex_attribute(vertex, "y", X[i + vcount].item())

        diff = norm(red_r)
        if n_iter > max_iter:
            raise SolutionError("Did not converge")

        print("i: {0:0} diff: {1:.2e}".format(n_iter, float(diff)))
        n_iter += 1

    print("Converged in {0} iterations".format(n_iter))

    return form


# ==============================================================================
# update force diagram
# ==============================================================================


def force_update_from_form(force: ForceDiagram, form: FormDiagram) -> ForceDiagram:
    """Update the force diagram after modifying the (force densities of) the form diagram.

    Parameters
    ----------
    force : :class:`ForceDiagram`
        The force diagram on which the update is based.
    form : :class:`FormDiagram`
        The form diagram to update.

    Returns
    -------
    force: :class:`ForceDiagram`
        The updated force diagram.

    """
    # --------------------------------------------------------------------------
    # form diagram
    # --------------------------------------------------------------------------
    vertex_index = form.vertex_index()

    xy = array(form.xy(), dtype=float64)
    edges = [[vertex_index[u], vertex_index[v]] for u, v in form.edges_where({"_is_edge": True})]
    C = connectivity_matrix(edges, "csr")
    Q = diags([form.q()], [0])
    uv = C.dot(xy)
    # --------------------------------------------------------------------------
    # force diagram
    # --------------------------------------------------------------------------
    _vertex_index = force.vertex_index()

    _known = [_vertex_index[force.anchor()]]
    _xy = array(force.xy(), dtype=float64)
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]
    _C = connectivity_matrix(_edges, "csr")
    _Ct = _C.transpose()
    # --------------------------------------------------------------------------
    # compute reciprocal for given q
    # --------------------------------------------------------------------------
    _xy = spsolve_with_known(_Ct.dot(_C), _Ct.dot(Q).dot(uv), _xy, _known)
    # --------------------------------------------------------------------------
    # update force diagram
    # --------------------------------------------------------------------------
    for vertex, attr in force.vertices(True):
        index = _vertex_index[vertex]
        attr["x"] = _xy[index, 0]
        attr["y"] = _xy[index, 1]

    return force


def force_update_from_form_geometrical(force: ForceDiagram, form: FormDiagram, kmax: int = 100) -> ForceDiagram:
    """Update the force diagram after modifying the (geometry of) the form diagram.

    Parameters
    ----------
    force : :class:`ForceDiagram`
        The force diagram on which the update is based.
    form : :class:`FormDiagram`
        The form diagram to update.
    kmax: int, optional
        Maximum number of least-square iterations for solving the duality form-force.
        The default value is ``20``.

    Returns
    -------
    force :class:`ForceDiagram`
        The updated force diagram.

    """
    # --------------------------------------------------------------------------
    # form diagram
    # --------------------------------------------------------------------------
    vertex_index = form.vertex_index()

    xy = array(form.xy(), dtype=float64)
    edges = [(vertex_index[u], vertex_index[v]) for u, v in form.edges()]
    C = connectivity_matrix(edges, "csr")

    # --------------------------------------------------------------------------
    # force diagram
    # --------------------------------------------------------------------------
    _vertex_index = force.vertex_index()
    _edge_index = force.edge_index(form)
    _edge_index.update({(v, u): _edge_index[u, v] for u, v in _edge_index})

    _xy = array(force.xy(), dtype=float64)
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]

    _i_j = {index: [_vertex_index[nbr] for nbr in force.vertex_neighbors(vertex)] for index, vertex in enumerate(force.vertices())}
    _ij_e = {(_vertex_index[u], _vertex_index[v]): _edge_index[u, v] for u, v in _edge_index}
    _ij_e.update({(_vertex_index[v], _vertex_index[u]): _edge_index[u, v] for u, v in _edge_index})

    # --------------------------------------------------------------------------
    # constraints
    # --------------------------------------------------------------------------
    _fixed = [_vertex_index[vertex] for vertex in force.fixed()]
    _free = list(set(range(force.number_of_vertices())) - set(_fixed))
    # _fixed_x = [_vertex_index[vertex] for vertex in force.fixed_x()]
    # _fixed_y = [_vertex_index[vertex] for vertex in force.fixed_y()]
    _line_constraints_all = force.vertices_attribute("line_constraint")
    _line_constraints = [_line_constraints_all[i] for i in _free]
    _target_lengths = form.edges_attribute("target_force")
    _target_vectors = [force.edge_attribute(edge, "target_vector") for edge in force.ordered_edges(form)]

    # --------------------------------------------------------------------------
    # compute the coordinates of the *free* vertices of the force diagram
    # as a function of the fixed vertices and the previous coordinates of the *free* vertices
    # --------------------------------------------------------------------------
    update_primal_from_dual(
        _xy,
        xy,
        _free,
        _i_j,
        _ij_e,
        C,
        line_constraints=_line_constraints,
        target_lengths=_target_lengths,
        target_vectors=_target_vectors,
        kmax=kmax,
    )

    # --------------------------------------------------------------------------
    # update force diagram
    # --------------------------------------------------------------------------
    for vertex, attr in force.vertices(True):
        index = _vertex_index[vertex]
        attr["x"] = _xy[index, 0]
        attr["y"] = _xy[index, 1]

    return force


def force_update_from_constraints(force: ForceDiagram, kmax: int = 100) -> ForceDiagram:
    """Update the force diagram from constraints on length and orientation imposed in the form diagram,
    and already carried out as attributes in the force diagram.

    Parameters
    ----------
    force : :class:`ForceDiagram`
        The force diagram on which the update is based.
    kmax: int, optional
        Maximum number of parallelisation iterations for updating the force diagram.
        The default value is ``100``.

    Returns
    -------
    force :class:`ForceDiagram`
        The updated force diagram.

    """

    # --------------------------------------------------------------------------
    # parameters from force diagram
    # --------------------------------------------------------------------------
    _k_i = force.vertex_index()
    _xy = force.vertices_attributes("xy")
    _edges = list(force.edges())
    _edges = [(_k_i[u], _k_i[v]) for u, v in _edges]
    _i_nbrs = {_k_i[key]: [_k_i[nbr] for nbr in force.vertex_neighbors(key)] for key in force.vertices()}
    _uv_i = {uv: index for index, uv in enumerate(_edges)}
    _ij_e = {(u, v): index for (u, v), index in iter(_uv_i.items())}

    # --------------------------------------------------------------------------
    # fixity from constraints on force diagram
    # --------------------------------------------------------------------------
    fixed_force = [key for key in force.vertices_where({"is_fixed": True})]
    _fixed = [_k_i[key] for key in fixed_force]
    line_constraints = force.vertices_attribute("line_constraint")

    # --------------------------------------------------------------------------
    # edge orientations and edge target lengths
    # --------------------------------------------------------------------------
    target_lengths = [force.dual_edge_targetforce(edge) for edge in force.edges()]
    target_vectors = force.edges_attribute("target_vector")

    # --------------------------------------------------------------------------
    # Paralelise edge given force targets and/or target_lengths
    # --------------------------------------------------------------------------
    parallelise_edges(
        _xy,
        _edges,
        _i_nbrs,
        _ij_e,
        target_vectors,
        target_lengths,
        fixed=_fixed,
        line_constraints=line_constraints,
        kmax=kmax,
    )

    # --------------------------------------------------------------------------
    # update force diagram geometry
    # --------------------------------------------------------------------------
    for key in force.vertices():
        i = _k_i[key]
        x, y = _xy[i]
        force.vertex_attribute(key, "x", x)
        force.vertex_attribute(key, "y", y)

    return force


# ==============================================================================
# dual modifications
# ==============================================================================


def update_diagrams_from_constraints(
    form: FormDiagram,
    force: ForceDiagram,
    max_iter: int = 20,
    kmax: int = 20,
    callback: Callable = None,
) -> tuple[FormDiagram, ForceDiagram]:
    """Update the form and force diagram after constraints / or movements are imposed to the diagrams.

    Parameters
    ----------
    form : :class:`FormDiagram`
        The form diagram.
    force : :class:`ForceDiagram`
        The force diagram.
    max_iter: int, optional
        Maximum number of iterations to update the diagrams.
        The default value is ``20``.
    kmax: int, optional
        Maximum number of least-square iterations for solving the duality form-force.
        The default value is ``20``.
    callback: callable, optional
        Callable function at the end of each iteration.
        The default value is ``None``.

    Returns
    -------
    form: :class:`FormDiagram`
        The updated form diagram.
    force: :class:`ForceDiagram`
        The updated force diagram.

    """

    form.dual = force
    force.dual = form

    niter = 0

    while niter < max_iter:
        # Propose a force diagram based on constraints -> Using paralellise
        force_update_from_constraints(force)

        if callback:
            callback(form, force)

        # Find geometrical dual form diagram respecting form constraints -> Using Least-Squares
        form_update_from_force(form, force, kmax=kmax)

        if callback:
            callback(form, force)

        # Find geometrical dual force diagram respecting force constraints -> Using Least-Squares
        force_update_from_form_geometrical(force, form, kmax=kmax)

        if callback:
            callback(form, force)

        niter += 1

    return form, force



from typing import Generator
from typing import Optional
from typing import Union

from compas.datastructures.mesh.duality import mesh_dual
from compas.geometry import Line
from compas_ags.diagrams import Diagram
from compas_ags.diagrams import FormDiagram  # noqa: F401


class ForceDiagram(Diagram):
    """Mesh-based data structure for force diagrams in AGS."""

    dual: FormDiagram

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.update_default_vertex_attributes(
            is_fixed=False,
            line_constraint=None,
            is_param=False,
        )
        self.update_default_edge_attributes(
            l=0.0,
            target_vector=None,
        )

    # --------------------------------------------------------------------------
    # Constructors
    # --------------------------------------------------------------------------

    @classmethod
    def from_formdiagram(cls, formdiagram: FormDiagram) -> "ForceDiagram":
        """Construct a force diagram from a form diagram.

        Parameters
        ----------
        formdiagram : :class:`compas_tna.diagrams.FormDiagram`
            The form diagram.

        Returns
        -------
        :class:`compas_ags.diagrams.ForceDiagram`

        """
        forcediagram: ForceDiagram = mesh_dual(formdiagram, cls)
        forcediagram.dual = formdiagram
        formdiagram.dual = forcediagram
        return forcediagram

    # --------------------------------------------------------------------------
    # Convenience functions for retrieving attributes of the force diagram.
    # --------------------------------------------------------------------------

    def xy(self) -> list[list[float]]:
        """The XY coordinates of the vertices.

        Returns
        -------
        list

        """
        return self.vertices_attributes("xy")

    def fixed(self) -> list[int]:
        """The identifiers of the fixed vertices.

        Returns
        -------
        list

        """
        return list(self.vertices_where({"is_fixed": True}))

    def anchor(self) -> list[int]:
        """Get an anchor to the force diagram.

        Returns
        -------
        int

        """
        return next(self.vertices())

    # --------------------------------------------------------------------------
    # Helpers
    # --------------------------------------------------------------------------

    def edges_where_dual(
        self,
        conditions: dict,
        data: bool = False,
    ) -> Generator[Union[tuple[int, int], tuple[tuple[int, int], dict]], None, None]:
        """Get edges for which a certain condition or set of conditions is true for the corresponding edges in the diagram's dual.

        Parameters
        ----------
        conditions : dict
            A set of conditions in the form of key-value pairs.
            The keys should be attribute names. The values can be attribute
            values or ranges of attribute values in the form of min/max pairs.
        data : bool, optional
            Yield the edges and their data attributes.
            Default is ``False``.

        Yields
        ------
        2-tuple
            The next edge as a (u, v) tuple, if ``data=False``.
            The next edge as a ((u, v), data) tuple, if ``data=True``.

        """
        for edge in list(self.edges()):
            is_match = True

            dual_edge = self.dual_edge(edge)
            dual_edge_attr = self.dual.edge_attributes(dual_edge)

            for cond_name, cond_value in conditions.items():
                dual_method = getattr(self.dual, cond_name, None)

                if dual_method and callable(dual_method):
                    dual_value = dual_method(dual_edge)
                elif cond_name in dual_edge_attr:
                    dual_value = dual_edge_attr[cond_name]
                else:
                    is_match = False
                    break

                if isinstance(dual_value, list):
                    if cond_value not in dual_value:
                        is_match = False
                        break
                elif isinstance(cond_value, (tuple, list)):
                    minval, maxval = cond_value
                    if dual_value < minval or dual_value > maxval:
                        is_match = False
                        break
                else:
                    if cond_value != dual_value:
                        is_match = False
                        break

            if is_match:
                if data:
                    yield edge, self.edge_attributes(edge)
                else:
                    yield edge

    def dual_edge(self, edge: tuple[int, int]) -> Union[tuple[int, int], None]:
        """Find the cooresponding edge in the diagram's dual.

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        tuple (int, int) or None
            The identifier of the dual edge if it exists.

        """
        for u, v in self.dual.face_halfedges(edge[0]):
            if self.dual.halfedge[v][u] == edge[1]:
                if self.dual.has_edge((u, v)):
                    return u, v
                return v, u

    def is_dual_edge_external(self, edge: tuple[int, int]) -> bool:
        """Verify if the corresponding edge in the diagram's dual is marked as "external".

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        bool

        """
        return self.dual.edge_attribute(self.dual_edge(edge), "is_external")

    def is_dual_edge_reaction(self, edge: tuple[int, int]) -> bool:
        """Verify if the corresponding edge in the diagram's dual is marked as "reaction".

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        bool

        """
        return self.dual.edge_attribute(self.dual_edge(edge), "is_reaction")

    def is_dual_edge_load(self, edge: tuple[int, int]) -> bool:
        """Verify if the corresponding edge in the diagram's dual is marked as "load".

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        bool

        """
        return self.dual.edge_attribute(self.dual_edge(edge), "is_load")

    def is_dual_edge_ind(self, edge: tuple[int, int]) -> bool:
        """Verify if the corresponding edge in the diagram's dual is marked as "independent".

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        bool

        """
        return self.dual.edge_attribute(self.dual_edge(edge), "is_ind")

    def dual_edge_force(self, edge: tuple[int, int]) -> float:
        """Retrieve the force in the corresponding edge of the diagram's dual.

        Parameters
        ----------
        edge : tuple(int, int)
            The edge identifier.

        Returns
        -------
        float

        """
        return self.dual.edge_attribute(self.dual_edge(edge), "f")

    def dual_edge_angledeviation(self, edge: tuple[int, int]) -> float:
        """Retrieve the angle deviation in the corresponding edge of the diagram's dual.

        Parameters
        ----------
        edge : tuple(int, int)
            The edge identifier.

        Returns
        -------
        float

        """
        return self.dual.edge_attribute(self.dual_edge(edge), "a")

    def dual_edge_targetforce(self, edge):
        """Retrieve the target force in the corresponding edge of the diagram's dual.

        Parameters
        ----------
        edge : tuple(int, int)
            The edge identifier.

        Returns
        -------
        float
        """
        return self.dual.edge_attribute(self.dual_edge(edge), "target_force")

    def edge_index(self, form: Optional[FormDiagram] = None) -> dict[tuple[int, int], int]:
        """Construct a mapping between the identifiers of edges and the corresponding indices in a list of edges.

        Parameters
        ----------
        form : :class:`compas_ags.diagrams.FormDiagram`, optional
            If a form diagra is provided as reference, the list of edges is ordered such that it corresponds
            to the natural ordering of edges in the form diagram.

        Returns
        -------
        dict
            Mapping between edge identifiers and the correpsonding indices of the edges in a list.

        """
        if not form:
            return {edge: index for index, edge in enumerate(self.edges())}
        edge_index = dict()
        for index, (u, v) in enumerate(form.edges()):
            f1 = form.halfedge[u][v]
            f2 = form.halfedge[v][u]
            edge_index[f1, f2] = index
            # the weird side-effect of this is that edges get rotated if necessary
        return edge_index

    def ordered_edges(self, form: FormDiagram) -> list[tuple[int, int]]:
        """ "Construct a list of edges with the same order as the corresponding edges of the form diagram.

        Parameters
        ----------
        form : :class:`compas_ags.diagrams.FormDiagram`

        Returns
        -------
        list
        """
        edge_index = self.edge_index(form=form)
        index_edge = {index: edge for edge, index in edge_index.items()}
        edges = [index_edge[index] for index in range(self.number_of_edges())]
        return edges

    # --------------------------------------------------------------------------
    # Helpers
    # --------------------------------------------------------------------------

    def constraints_from_dual(self, tol: float = 10e-4) -> None:
        """ "Reflect constraints from the form diagram in the force diagram."""
        edge_index = self.dual.edge_index()
        ordered_edges = self.ordered_edges(self.dual)
        edges_orient = []

        # Fix vertices of dual independent edge
        for edge in self.edges_where_dual({"is_ind": True}):
            self.vertices_attribute("is_fixed", True, keys=edge)
            edges_orient.append(edge)

        # If loads are orthogonal the force dual edge gets constrained
        for edge in self.edges_where_dual({"is_load": True}):
            self.edge_attribute(edge, "is_load", True)
            edges_orient.append(edge)
            sp, ep = self.edge_coordinates(edge)
            line = Line(sp, ep)
            self.vertices_attribute("line_constraint", value=line, keys=edge)

        for edge in self.edges_where_dual({"is_reaction": True}):
            self.edge_attribute(edge, "is_reaction", True)
            edges_orient.append(edge)

        for form_edge in self.dual.edges():
            target_vector = self.dual.edge_attribute(form_edge, "target_vector")
            index = edge_index[form_edge]
            force_edge = ordered_edges[index]
            if target_vector is not None:
                edges_orient.append(force_edge)

        for edge in edges_orient:
            edge = edge if edge in list(self.edges()) else (edge[1], edge[0])
            sp, ep = self.edge_coordinates(edge)
            dx = ep[0] - sp[0]
            dy = ep[1] - sp[1]
            length = (dx**2 + dy**2) ** 0.5
            self.edge_attribute(edge, "target_vector", [dx / length, dy / length])

    # def compute_constraints(self, form, M):
    #     r"""Computes the form diagram constraints used
    #     in compas_bi_ags.bi_ags.graphstatics.form_update_from_force_direct

    #     Parameters
    #     ----------
    #     form : compas_ags.diagrams.formdiagram.FormDiagram
    #         The form diagram to update.
    #     M
    #         The matrix described in compas_bi_ags.bi_ags.graphstatics.form_update_from_force_direct
    #     """
    #     import numpy as np
    #     nr_col_jac = M.shape[1]
    #     constraint_rows = np.zeros((0, M.shape[1]))
    #     residual = np.zeros((0, 1))
    #     vcount = form.number_of_vertices()

    #     # Currently this computes two constraints per fixed vertex in the form diagram.
    #     for i, (key, attr) in enumerate(form.vertices(True)):
    #         if not attr['is_fixed']:
    #             continue

    #         # Handle x
    #         constraint_jac_row = np.zeros(
    #             (1, nr_col_jac))  # Added row for jacobian
    #         # Lock horizontal position
    #         constraint_jac_row[0, i] = 1
    #         constraint_rows = np.vstack((constraint_rows, constraint_jac_row))
    #         residual = np.vstack((residual, attr['x']))

    #         # Handle y
    #         constraint_jac_row = np.zeros(
    #             (1, nr_col_jac))  # Added row for jacobian
    #         # Lock horizontal position
    #         constraint_jac_row[0, i+vcount] = 1
    #         constraint_rows = np.vstack((constraint_rows, constraint_jac_row))
    #         residual = np.vstack((residual, attr['y']))
    #     return constraint_rows, residual



import os

HERE = os.path.abspath(os.path.dirname(__file__))
HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.join(HOME, "data")
TEMP = os.path.join(HOME, "temp")


__author__ = "Tom Van Mele and others"
__copyright__ = "Copyright 2014-2018 - Block Research Group, ETH Zurich"
__license__ = "MIT License"
__email__ = "vanmelet@ethz.ch"
__version__ = "1.3.1"


def get(relpath):
    return os.path.join(DATA, relpath)


__all__ = ["HOME", "DATA", "DOCS", "TEMP"]



from compas_viewer import Viewer
from compas_viewer.config import Config

from compas.colors import Color
from compas.geometry import Box
from compas.geometry import Circle
from compas.geometry import Polygon
from compas.geometry import bounding_box
from compas_ags.diagrams import ForceDiagram
from compas_ags.diagrams import FormDiagram

loadcolor = Color.green().darkened(50)
reactioncolor = Color.green().darkened(50)
tensioncolor = Color.red().lightened(25)
compressioncolor = Color.blue().lightened(25)


class AGSViewer(Viewer):
    def __init__(self, config: Config | None = None, **kwargs):
        config = config or Config()
        config.renderer.view = "top"
        config.camera.target = [0, 0, 0]
        config.camera.position = [0, 0, 10]
        config.renderer.gridsize = [100, 100, 100, 100]

        super().__init__(config, **kwargs)

        self.renderer.camera.rotation.set(0, 0, 0, False)

        self.form: FormDiagram = None
        self.force: ForceDiagram = None

    def add_form(self, form: FormDiagram, show_forces=True, scale_forces=0.1, nodesize=0.1):
        self.form = form

        self.scene.add(
            self.form,
            show_faces=False,
            show_lines=not show_forces,
            linewidth=2,
            name="FormDiagram",
        )

        nodes = []
        for vertex in self.form.vertices():
            circle = Circle.from_point_and_radius(
                point=self.form.vertex_point(vertex) + [0, 0, 0.001],
                radius=nodesize,
            )
            nodes.append(circle.to_polygon(n=128))

        self.scene.add(
            nodes,
            name="Nodes",
            facecolor=Color.white(),
            linecolor=Color.black(),
        )

        if not show_forces:
            return

        external = []
        compression = []
        tension = []
        for edge in self.form.edges():
            line = self.form.edge_line(edge)
            vector = line.direction.cross([0, 0, 1])
            force = self.form.edge_attribute(edge, name="f")
            is_external = self.form.edge_attribute(edge, name="is_external")

            w = scale_forces * 0.5 * abs(force)
            a = line.start + vector * -w
            b = line.end + vector * -w
            c = line.end + vector * +w
            d = line.start + vector * +w

            if is_external:
                external.append(Polygon([a, b, c, d]))
            elif force > 0:
                tension.append(Polygon([a, b, c, d]))
            elif force < 0:
                compression.append(Polygon([a, b, c, d]))

        self.scene.add(
            external,
            name="External Forces",
            facecolor=reactioncolor,
            linecolor=reactioncolor.contrast,
        )
        self.scene.add(
            compression,
            name="Compression",
            facecolor=compressioncolor,
            linecolor=compressioncolor.contrast,
        )
        self.scene.add(
            tension,
            name="Tension",
            facecolor=tensioncolor,
            linecolor=tensioncolor.contrast,
        )

    def add_force(self, force: ForceDiagram):
        self.force = force

        b1 = Box.from_bounding_box(bounding_box(self.form.vertices_attributes("xyz")))
        b2 = Box.from_bounding_box(bounding_box(self.force.vertices_attributes("xyz")))

        dx = b2.xmin - b1.xmax
        if dx < 1:
            dx = 1.5 * (b1.xmax - b2.xmin)
        else:
            dx = 0

        self.scene.add(
            force.translated([dx, 0, 0]),
            show_faces=False,
            linewidth=2,
            name="ForceDiagram",
        )



import math
from abc import ABC
from abc import abstractmethod

import numpy as np
import numpy.typing as npt

from compas_ags.diagrams import FormDiagram


class AbstractConstraint(ABC):
    """Base class for Form Diagram constraints.

    Parameters
    -----------
    form : :class:`FormDiagram`

    """

    def __init__(self, form: FormDiagram) -> None:
        super().__init__()
        self.form = form
        self.vcount = form.number_of_vertices()
        self.vertex_index = form.vertex_index()
        self._color = (21, 36, 198)
        self._width = 1.0
        self._style = "--"

    @abstractmethod
    def compute_constraint(self):
        """Computes the residual and Jacobian matrix of the constraint."""
        pass

    @abstractmethod
    def update_constraint_goal(self):
        """Update constraint values based on current form diagram"""
        pass

    @property
    def number_of_cols(self):
        vcount = self.form.number_of_vertices()
        return 2 * vcount

    def get_lines(self):
        """Get lines to draw in viewer."""
        pass


class ConstraintsCollection:
    """Computes the Jacobian d_X/dX and residual r of the added constraints
    where X contains the form diagram coordinates in *Fortran* order
    (first all x-coordinates, then all y-coordinates) and _X contains the
    force diagram coordinates in *Fortran* order (first all _x-coordinates,
    then all _y-coordinates).

    Parameters
    -----------
    form : :class:`FormDiagram`

    References
    ----------
    .. [1] Alic, V. and Åkesson, D., 2017. Bi-directional algebraic graphic statics. Computer-Aided Design, 93, pp.26-37.

    """

    def __init__(self, form: FormDiagram) -> None:
        self.constraints: list[AbstractConstraint] = []
        self.form = form

    def add_constraint(self, constraint: AbstractConstraint) -> None:
        self.constraints.append(constraint)

    def compute_constraints(self) -> tuple[npt.NDArray, npt.NDArray]:
        jac = np.zeros((0, self.constraints[0].number_of_cols))
        res = np.zeros((0, 1))
        for constraint in self.constraints:
            j, r = constraint.compute_constraint()
            jac = np.vstack((jac, j))
            res = np.vstack((res, r))
        return jac, res

    def update_constraints(self) -> None:
        for constraint in self.constraints:
            constraint.update_constraint_goal()

    def constraints_from_form(self) -> None:
        # fix x and y coordinates of the fixed vertices
        for key in self.form.vertices_where(is_fixed=True):
            self.add_constraint(HorizontalFix(self.form, key))
            self.add_constraint(VerticalFix(self.form, key))

        # fix x or y coordinates of the non-fixed vertices
        for key in self.form.vertices_where(is_fixed_x=True, is_fixed=False):
            self.add_constraint(HorizontalFix(self.form, key))
        for key in self.form.vertices_where(is_fixed_y=True, is_fixed=False):
            self.add_constraint(VerticalFix(self.form, key))

        self.constrain_dependent_leaf_edges_lengths()

    def get_lines(self) -> list[dict]:
        lines = []
        for constraint in self.constraints:
            constraint_lines = constraint.get_lines()
            if constraint_lines:
                lines = lines + constraint_lines
        return lines

    def constrain_dependent_leaf_edges_lengths(self) -> None:
        leaves = self.form.leaves()
        dependent_leaf_edges = []
        for i, (u, v) in enumerate(self.form.edges()):
            if u in leaves or v in leaves:
                if not self.form.edge_attribute((u, v), name="is_ind"):
                    dependent_leaf_edges.append((u, v))
        for edge in dependent_leaf_edges:
            self.add_constraint(LengthFix(self.form, edge))


class HorizontalFix(AbstractConstraint):
    """Constraint that keeps x-coordinate of a vertex fixed.

    Parameters
    -----------
    form : :class:`FormDiagram`
        The Form Diagram to constraint
    vertex : int
        Idenfifier ofthe vertex to fix.

    """

    def __init__(self, form: FormDiagram, vertex: int) -> None:
        super().__init__(form)
        self.vertex = vertex
        self.x = None
        self.set_initial_position()

    def set_initial_position(self) -> None:
        self.x = self.form.vertex_attribute(self.vertex, "x")

    def compute_constraint(self) -> tuple[npt.NDArray, float]:
        constraint_jac_row = np.zeros((1, self.number_of_cols))
        idx = self.vertex_index[self.vertex]
        constraint_jac_row[0, idx] = 1
        r = self.form.vertex_attribute(self.vertex, "x") - self.x
        return constraint_jac_row, r

    def update_constraint_goal(self) -> None:
        self.set_initial_position()

    def get_lines(self) -> list[dict]:
        constraint_lines = []
        s = self.form.vertex_coordinates(self.vertex, "xy")
        e = self.form.vertex_coordinates(self.vertex, "xy")
        s[1] += 1
        e[1] -= 1
        constraint_lines.append(
            {
                "start": s,
                "end": e,
                "width": self._width,
                "color": self._color,
                "style": self._style,
            }
        )
        return constraint_lines


class VerticalFix(AbstractConstraint):
    """Constraint that keeps y-coordinate of a vertex fixed.

    Parameters
    -----------
    form : :class:`FormDiagram`
        The Form Diagram to constraint
    vertex : int
        Idenfifier ofthe vertex to fix.

    """

    def __init__(self, form: FormDiagram, vertex: int):
        super().__init__(form)
        self.vertex = vertex
        self.y = None
        self.set_initial_position()

    def set_initial_position(self) -> None:
        self.y = self.form.vertex_attribute(self.vertex, "y")

    def compute_constraint(self) -> tuple[npt.NDArray, float]:
        constraint_jac_row = np.zeros((1, self.number_of_cols))
        idx = self.vertex_index[self.vertex] + self.vcount
        constraint_jac_row[0, idx] = 1
        r = self.form.vertex_attribute(self.vertex, "y") - self.y
        return constraint_jac_row, r

    def update_constraint_goal(self) -> None:
        self.set_initial_position()

    def get_lines(self) -> list[dict]:
        constraint_lines = []
        s = self.form.vertex_coordinates(self.vertex, "xy")
        e = self.form.vertex_coordinates(self.vertex, "xy")
        s[0] += 1
        e[0] -= 1
        constraint_lines.append(
            {
                "start": s,
                "end": e,
                "width": self._width,
                "color": self._color,
                "style": self._style,
            }
        )
        return constraint_lines


class AngleFix(AbstractConstraint):
    """Constraint that keeps vertex fixed along an inclined line.

    Parameters
    -----------
    form : :class:`FormDiagram`
        The Form Diagram to constraint
    vertex : int
        Idenfifier ofthe vertex to fix.
    angle : float
        Angle (clockwise) to fix the vertex to.

    """

    def __init__(self, form: FormDiagram, vertex: int, angle: float) -> None:
        super().__init__(form)
        self.vertex = vertex
        self.angle = angle
        self.x = None
        self.y = None
        self.set_initial_position()

    def set_initial_position(self) -> None:
        self.x = self.form.vertex_attribute(self.vertex, "x")
        self.y = self.form.vertex_attribute(self.vertex, "y")

    def compute_constraint(self) -> tuple[npt.NDArray, float]:
        constraint_jac_row = np.zeros((1, self.number_of_cols))

        theta = math.radians(self.angle)

        idx = self.vertex_index[self.vertex]
        constraint_jac_row[0, idx] = 1 * math.sin(theta)
        r = (self.form.vertex_attribute(self.vertex, "x") - self.x) * math.sin(theta)

        idx = self.vertex_index[self.vertex] + self.vcount
        constraint_jac_row[0, idx] = 1 * math.cos(theta)
        r = (self.form.vertex_attribute(self.vertex, "y") - self.y) * math.cos(theta)
        return constraint_jac_row, r

    def update_constraint_goal(self) -> None:
        self.set_initial_position()

    def get_lines(self) -> list[dict]:
        constraint_lines = []
        s = self.form.vertex_coordinates(self.vertex, "xy")
        e = self.form.vertex_coordinates(self.vertex, "xy")
        theta = math.radians(90 - self.angle)
        s[0] += 1 * math.sin(theta)
        s[1] -= 1 * math.cos(theta)
        e[0] -= 1 * math.sin(theta)
        e[1] += 1 * math.cos(theta)
        constraint_lines.append(
            {
                "start": s,
                "end": e,
                "width": self._width,
                "color": self._color,
                "style": self._style,
            }
        )
        return constraint_lines


class LengthFix(AbstractConstraint):
    """Constraint that keeps the edge length fixed.

    Parameters
    -----------
    form : :class:`FormDiagram`
        The Form Diagram to constraint
    edge : 2-tuple of int
        The identifier of the edge as a pair of vertex identifiers.

    """

    def __init__(self, form: FormDiagram, edge: tuple[int, int]):
        super().__init__(form)
        self.edge = edge
        self.length = None
        self.set_initial_length()

    def update_constraint_goal(self) -> None:
        self.set_initial_length()

    def set_initial_length(self) -> None:
        self.length = self.form.edge_length(*self.edge)  # Initial length

    def compute_constraint(self) -> tuple[npt.NDArray, float]:
        constraint_jac_row = np.zeros((1, self.number_of_cols))

        s, e = self.form.edge_coordinates(*self.edge)
        dx = s[0] - e[0]
        dy = s[1] - e[1]
        length = math.sqrt(dx**2 + dy**2)  # Current length

        id_u = self.vertex_index[self.edge[0]]
        id_v = self.vertex_index[self.edge[1]]

        constraint_jac_row[0, id_u] = dx / length  # x0
        constraint_jac_row[0, id_v] = -dx / length  # x1
        constraint_jac_row[0, id_u + self.vcount] = dy / length  # y0
        constraint_jac_row[0, id_v + self.vcount] = -dy / length  # y1
        r = length - self.length

        return constraint_jac_row, r


class SetLength(LengthFix):
    """WIP - Constraint that sets the edge length to L.

    Parameters
    -----------
    form : :class:`FormDiagram`
        The Form Diagram to constraint
    edge : 2-tuple of int
        The identifier of the edge as a pair of vertex identifiers.
    length : float
        Length to set to the edge.

    """

    def __init__(self, form: FormDiagram, edge: tuple[int, int], length: float):
        super().__init__(form, edge)
        self.length = length

    def update_constraint_goal(self) -> None:
        pass


class SetOrientation(AbstractConstraint):
    """WIP - Constraint that sets the edge orientation.

    Parameters
    -----------
    form : :class:`FormDiagram`
        The Form Diagram to constraint
    edge : 2-tuple of int
        The identifier of the edge as a pair of vertex identifiers.
    angle : float
        Angle to orient the edge to set to the edge.

    """

    def __init__(self, form: FormDiagram, edge: tuple[int, int], angle: float):
        super().__init__(form)
        self.edge = edge
        self.angle = angle



from itertools import islice
from typing import Generator
from typing import Optional
from typing import Union

from compas.geometry import Line
from compas_ags.diagrams import Diagram
from compas_ags.diagrams import FormGraph


class FormDiagram(Diagram):
    """Mesh-based data structure for form diagrams in AGS."""

    def __init__(self, **kwargs):
        super(FormDiagram, self).__init__(**kwargs)
        self._graph = None
        self.attributes.update(
            {
                "name": "Form",
            }
        )
        self.update_default_vertex_attributes(
            {
                "is_fixed": False,
                "line_constraint": None,
                "cx": 0.0,
                "cy": 0.0,
            }
        )
        self.update_default_edge_attributes(
            {
                "_is_edge": True,
                "a": 0.0,
                "q": 1.0,
                "f": 0.0,
                "l": 0.0,
                "is_ind": False,
                "is_external": False,
                "is_reaction": False,
                "is_load": False,
                "target_vector": None,
                "target_force": None,
            }
        )

    @property
    def graph(self) -> FormGraph:
        return self._graph

    @graph.setter
    def graph(self, graph: FormGraph):
        self._graph = graph

    @classmethod
    def from_graph(cls, graph: FormGraph) -> "FormDiagram":
        """Construct a form diagram from a form graph.

        This constructor converts the form graph into a mesh by finding the cycles of its planar embedding.
        Note that tt doesn't check if the graph actually is a planar embedding.
        The outside face of the mesh is automatically split into smaller faces at the leaves.

        Parameters
        ----------
        graph : :class:`compas_ags.diagrams.FormGraph`

        Returns
        -------
        :class:`compas_ags.diagrams.FormDiagram`
        """
        for node in list(graph.nodes()):
            if graph.degree(node) == 2:
                graph.delete_node(node)
        node_index = graph.node_index()
        cycles = graph.find_cycles(breakpoints=graph.leaves())
        points = graph.nodes_attributes("xyz")
        cycles[:] = [[node_index[node] for node in cycle] for cycle in cycles]
        form = cls.from_vertices_and_faces(points, cycles)
        form.edges_attribute("_is_edge", False, keys=list(form.edges_on_boundary()))
        form.edges_attribute("is_external", True, keys=form.leaf_edges())
        form.graph = graph
        return form

    # --------------------------------------------------------------------------
    # vertices
    # --------------------------------------------------------------------------

    def leaves(self) -> list[int]:
        """Identify the leaves of the form diagram.

        Returns
        -------
        list
            The identifiers of vertices with only one connected edge.
        """
        keys = []
        for key in self.vertices():
            edges = 0
            nbrs = self.vertex_neighbors(key)
            for nbr in nbrs:
                if self.edge_attribute((key, nbr), "_is_edge"):
                    edges += 1
            if edges == 1:
                keys.append(key)
        return keys

    # --------------------------------------------------------------------------
    # edges
    # --------------------------------------------------------------------------

    def edges(self, data: bool = False) -> Generator[Union[tuple[int, int], tuple[tuple[int, int], dict]], None, None]:
        """Edge iterator automatically discarding mesh edges that are not relevant in AGS.

        Parameters
        ----------
        data : bool, optional
            If `True`, yield the data attributes of the edges together with their identifiers.

        Yields
        ------
        tuple
            If `data` is `False`, the tuple of vertices identifying the edge.
            Otherwise, a tuple with the pair of vertices and an attribute dict.
        """
        seen = set()
        for u in self.halfedge:
            for v in self.halfedge[u]:
                if (u, v) in seen or (v, u) in seen:
                    continue
                seen.add((u, v))
                seen.add((v, u))
                if not self.edge_attribute((u, v), "_is_edge"):
                    continue
                if not data:
                    yield u, v
                else:
                    yield (u, v), self.edge_attributes((u, v))

    def leaf_edges(self) -> list[tuple[int, int]]:
        """Identify the edges connecting leaf vertices to the diagram.

        Returns
        -------
        list
            The identifiers of the edges.
        """
        edges = []
        leaves = set(self.leaves())
        for u, v in self.edges():
            if u in leaves or v in leaves:
                edges.append((u, v))
        return edges

    def edge_forcedensity(
        self,
        edge: Union[tuple[int, int], int],
        q: Optional[float] = None,
    ) -> float:
        """Get or set the forcedensity in an edge.

        Parameters
        ----------
        edge : int or tuple
            The identifier of the edge.
            This can be the index in the edge list or a tuple of vertices.
        q : float, optional
            If no new value is given, the current forcedensity value will be returned.
            Otherwise the stored value is updated with the provided one.

        Returns
        -------
        float
            The value of the force density in the edge.

        """
        if isinstance(edge, int):
            edge = next(islice(self.edges(), edge, None))

        if q is None:
            return self.edge_attribute(edge, "q")

        return self.edge_attribute(edge, "q", q)

    def edge_force(self, edge: Union[tuple[int, int], int], force: Optional[float] = None) -> float:
        """Get or set the force in an edge.

        Parameters
        ----------
        edge : int or tuple
            The identifier of the edge.
            This can be the index in the edge list or a tuple of vertices.
        force : float, optional
            If no value is given, the current force value will be returned.
            Otherwise the stored value is updated.

        Returns
        -------
        float
            The current force in the edge.

        """
        if isinstance(edge, int):
            edge = next(islice(self.edges(), edge, None))

        length = self.edge_length(edge)
        q = self.edge_attribute(edge, "q")

        if force is None:
            return q * length

        self.edge_attribute(edge, "is_ind", True)
        self.edge_attribute(edge, "q", force / length)
        return force

    # --------------------------------------------------------------------------
    # Convenience functions for retrieving the attributes of the formdiagram.
    # --------------------------------------------------------------------------

    def q(self) -> list[float]:
        return self.edges_attribute("q")

    def xy(self) -> list[list[float]]:
        return self.vertices_attributes("xy")

    def fixed(self) -> list[int]:
        return list(self.vertices_where(is_fixed=True))

    def constrained(self) -> list[int]:
        return [vertex for vertex, attr in self.vertices(True) if attr["cx"] or attr["cy"]]

    def constraints(self) -> tuple[float, float]:
        cx = self.vertices_attribute("cx")
        cy = self.vertices_attribute("cy")
        return cx, cy

    def ind(self) -> list[tuple[int, int]]:
        return list(self.edges_where(is_ind=True))

    # --------------------------------------------------------------------------
    # Identify features of the formdiagram based on geometrical inputs.
    # --------------------------------------------------------------------------

    def identify_constraints(self, tol: float = 10e-4) -> None:
        """Identify constraints on the Form Diagram based on the geometry.
        External loads define a line-load which constraint vertices in x, or y.

        Parameters
        ----------
        tol : float, optional
            Tolerance to define if leaves lay in a vertical, or horizontal line.
            The default value is `10E-4`.

        Returns
        -------
        None
            The FormDiagram is modified in place.

        """
        fixed = self.fixed()
        leaves = self.leaves()

        for edge in self.leaf_edges():
            sp, ep = self.edge_coordinates(edge)
            line = Line(sp, ep)
            dx = ep[0] - sp[0]
            dy = ep[1] - sp[1]
            length = (dx**2 + dy**2) ** 0.5

            self.edge_attribute(edge, "target_vector", [dx / length, dy / length])
            # by default loads are leaves connected to non fixed vertices
            self.edge_attribute(edge, "is_load", True)

            if edge[0] in fixed or edge[1] in fixed:
                # by default reactions are leaves connected to fixed vertices
                self.edge_attribute(edge, "is_reaction", True)
                self.edge_attribute(edge, "is_load", False)
                continue

            if edge[0] in leaves:
                self.vertex_attribute(edge[1], "line_constraint", line)
            else:
                self.vertex_attribute(edge[0], "line_constraint", line)

    # def identify_fixed(self, points=None, fix_degree=1):
    #     for key, attr in self.vertices(True):
    #         attr['is_fixed'] = self.vertex_degree(key) <= fix_degree
    #     if points:
    #         xy_key = {}
    #         for key in self.vertices():
    #             gkey = geometric_key_xy(self.vertex_attributes(key, 'xy'))
    #             xy_key[gkey] = key
    #         for xy in points:
    #             gkey = geometric_key_xy(xy)
    #             if gkey in xy_key:
    #                 key = xy_key[gkey]
    #                 self.vertex[key]['is_fixed'] = True

    # def identify_constraints(self, points=None):
    #     if points:
    #         xy_key = {}
    #         for key in self.vertices():
    #             gkey = geometric_key_xy(self.vertex_attributes(key, 'xy'))
    #             xy_key[gkey] = key
    #         for xy in points:
    #             gkey = geometric_key_xy(xy)
    #             if gkey in xy_key:
    #                 key = xy_key[gkey]
    #                 self.vertex[key]['cx'] = 1.0
    #                 self.vertex[key]['cy'] = 1.0



from .formgraph import FormGraph
from .diagram import Diagram
from .formdiagram import FormDiagram
from .forcediagram import ForceDiagram

__all__ = [
    "FormGraph",
    "Diagram",
    "FormDiagram",
    "ForceDiagram",
]



class SolutionError(Exception):
    """Used to throw solution errors during form or force computations."""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)
